//#define	SERIAL_PRINT

#include <LiquidCrystal.h>

#define	TITLE			"LBCK04"
#define	VERSION		"1.0"
#define COPYRIGHT	"(C) 2023 JH4VAJ"

#define VREF        5.0   /* リファレンス電圧 */
//#define VREF        4.95   /* リファレンス電圧 */

// LED 1
#define PIN_CTRL_1    A4    /* 電圧設定ボリュームピン */
#define PIN_VE_1      A0    /* エミッタ電圧測定ピン */
#define PIN_LED_VF_1  A2    /* LED Vf測定ピン */
#define PIN_VOUT_1    6     /* 電圧出力ピン */
#define RLED_LO_1     1000.0 /* LED電流制限抵抗値: 低電流 */
#define RLED_HI_1     33.3  /* LED電流制限抵抗値: 高電流 */
#define RFET_1        1.0   /* FETのON抵抗値 */

// LED 2
#define PIN_CTRL_2    A5    /* 電圧設定ボリュームピン */
#define PIN_VE_2      A1    /* エミッタ電圧測定ピン */
#define PIN_LED_VF_2  A3    /* LED Vf測定ピン */
#define PIN_VOUT_2    5     /* 電圧出力ピン */
#define RLED_LO_2     1000.0 /* LED電流制限抵抗値: 低電流 */
#define RLED_HI_2     33.3  /* LED電流制限抵抗値: 高電流 */
#define RFET_2        1.0   /* FETのON抵抗値 */

#define PIN_IFLVL     2     /* Ifレベル検出ピン */

#define MIN_VOUT_LO 120     /* 最小出力電圧（analogWrite()の値（0～255） */
#define MIN_VOUT_HI 140     /* 最小出力電圧（analogWrite()の値（0～255） */

#define PIN_PRINT_SW_1  12  /* 表示形式切替えスイッチピン */
#define PIN_PRINT_SW_2  11  /* 表示形式切替えスイッチピン */


float combined_rled_hi_1 = 1.0/((1.0/RLED_LO_1) + (1.0/(RLED_HI_1 + RFET_1))); // 高電流時の合成抵抗
float combined_rled_hi_2 = 1.0/((1.0/RLED_LO_2) + (1.0/(RLED_HI_2 + RFET_2)));

float v_lsb = VREF / 1024.0;  // ADC 1ビットあたりの電圧

float min_vout_lo_100 = 255.0 / (float)(255 - MIN_VOUT_LO); // 下駄分のMIN_VOUTを引いて残りを100%とする
float min_vout_hi_100 = 255.0 / (float)(255 - MIN_VOUT_HI); 

float ve_1, ve_2;                 // Ve
float led_vf_1, led_vf_2;         // LED_Vf
float led_if_1, led_if_2;         // LED_If

const int rs = 10, en = 9, d4 = 8, d5 = 7, d6 = 4, d7 = 3;
LiquidCrystal lcd(rs, en, d4, d5, d6, d7);


void setup() {
  pinMode(PIN_IFLVL, INPUT);
  pinMode(PIN_PRINT_SW_1, INPUT_PULLUP);
  pinMode(PIN_PRINT_SW_2, INPUT_PULLUP);

#ifdef SERIAL_PRINT  
  Serial.begin(9600);
#endif

  lcd.begin(16, 2);
  lcd.clear();
	
	print_version();
}



void print_version() {
#ifdef SERIAL_PRINT
	Serial.print(TITLE);
	Serial.print(" Ver.");
	Serial.println(VERSION);
	Serial.println(COPYRIGHT);
#endif

	lcd.clear();
	lcd.print(TITLE);
	lcd.print(" Ver.");
	lcd.print(VERSION);
	lcd.setCursor(0, 1);
	lcd.print(COPYRIGHT);
	delay(1000);
}


/* LED点灯制御: VRに合せてLEDの明るさを変える */
void ctrl_led() {
  int   adc_ctrl;     // ADC値: 電圧設定ボリューム

  /* LED 1 */
  if (digitalRead(PIN_IFLVL) == LOW) {
    adc_ctrl = (int)((float)(analogRead(PIN_CTRL_1) >> 2) / min_vout_lo_100) + MIN_VOUT_LO;
  }
  else {
    adc_ctrl = (int)((float)(analogRead(PIN_CTRL_1) >> 2) / min_vout_hi_100) + MIN_VOUT_HI;
  }

  analogWrite(PIN_VOUT_1, adc_ctrl);
  
  /* LED 2 */
  if (digitalRead(PIN_IFLVL) == LOW) {
    adc_ctrl = (int)((float)(analogRead(PIN_CTRL_2) >> 2) / min_vout_lo_100) + MIN_VOUT_LO;
  }
  else {
    adc_ctrl = (int)((float)(analogRead(PIN_CTRL_2) >> 2) / min_vout_hi_100) + MIN_VOUT_HI;
  }
    
  analogWrite(PIN_VOUT_2, adc_ctrl);

  return;
}


/* Ve, Vf読取り、If計算 */
void read_ve_vf_if() {
  int   adc_ve_1, adc_ve_2;         // ADC値: エミッタ電圧（＝抵抗を含めた印加電圧）
  int   adc_led_vf_1, adc_led_vf_2; // ADC値: LEDのVf
  float rled;                       // LED電流制限抵抗値
  
  const int   ave_num = 4;   // 測定値読み取り回数（平均回数）
  int i;

  adc_ve_1 = 0;
  adc_ve_2 = 0;
  adc_led_vf_1 = 0;
  adc_led_vf_2 = 0;

  for (i = 0; i < ave_num; i++) {
    /* LED 1 */
    adc_ve_1 += analogRead(PIN_VE_1); 
    adc_led_vf_1 += analogRead(PIN_LED_VF_1);

    /* LED 2 */
    adc_ve_2 += analogRead(PIN_VE_2);
    adc_led_vf_2 += analogRead(PIN_LED_VF_2);
  }

  adc_ve_1 /= ave_num;
  adc_ve_2 /= ave_num;
  adc_led_vf_1 /= ave_num;
  adc_led_vf_2 /= ave_num;

  /* LED 1 */
  ve_1 = v_lsb * (float)adc_ve_1;
  led_vf_1 = v_lsb * (float)adc_led_vf_1;

  if (adc_ve_1 > adc_led_vf_1) {
    rled = (digitalRead(PIN_IFLVL) == LOW) ? RLED_LO_1 : combined_rled_hi_1;
    led_if_1 = ((v_lsb * (float)(adc_ve_1 - adc_led_vf_1)) / rled);
  }
  else {
    led_if_1 = 0.0;
  }

  /* LED 2 */
  ve_2 = v_lsb * (float)adc_ve_2;
  led_vf_2 = v_lsb * (float)adc_led_vf_2;

  if (adc_ve_2 > adc_led_vf_2) {
    rled = (digitalRead(PIN_IFLVL) == LOW) ? RLED_LO_2 : combined_rled_hi_2;
    led_if_2 = ((v_lsb * (float)(adc_ve_2 - adc_led_vf_2)) / rled);
  }
  else {
    led_if_2 = 0.0;
  }
}


/* Vf、If値の表示 */
void print_vf_if() {
  char  str_buf[8];  

  /* Vf */
  dtostrf(led_vf_1, 5, 2, str_buf);
  lcd.print("Vf ");
  lcd.print(str_buf);
  lcd.print(" ");
  dtostrf(led_vf_2, 5, 2, str_buf);
  lcd.print(str_buf);
  lcd.print("V");

  /* If */
  dtostrf((led_if_1 * 1000.0), 5, 2, str_buf);
  lcd.setCursor(0, 1);
  lcd.print("If ");
  lcd.print(str_buf);
  lcd.print(" ");
  dtostrf((led_if_2 * 1000.0), 5, 2, str_buf);
  lcd.print(str_buf);
  lcd.print("mA");
}



/* 抵抗値を適切な表示形式（文字列）に変換する。バッファサイズはチェックしないことに要注意。 */
void r_val_fmt(float r, char *p) {
  if (r < 1000.0) {
    dtostrf(r, 3, 0, p);
    *(p + 3) = 0xf4;  // Ω
    *(p + 4) = '\0';
  }
  else {
    r /= 1000.0;
    
    if (r < 10.0) {
      dtostrf(r, 3, 1, p);
    }
    else {
      dtostrf(r, 3, 0, p);
    }

    if (strncmp(p, "INF", 3) != 0) {
      *(p + 3) = 'k';
      *(p + 4) = '\0';
    }
  }
}


/* 電流制限抵抗値の計算と表示 */
void print_r_val(float led_vf, float led_if) {
  char str_buf[8];

  lcd.setCursor(2, 0);
  lcd.print("3:");
  if (led_vf < 3.0) {
    r_val_fmt(((3.0 - led_vf) / led_if), str_buf);
  }
  else {
    strcpy(str_buf, "---");
  }
  lcd.print(str_buf);

  lcd.setCursor(9, 0);
  lcd.print(" 5:");
  r_val_fmt(((5.0 - led_vf) / led_if), str_buf);
  lcd.print(str_buf);

  
  lcd.setCursor(2, 1);
  lcd.print("9:");
  r_val_fmt(((9.0 - led_vf) / led_if), str_buf);
  lcd.print(str_buf);

  lcd.setCursor(9, 1);
  lcd.print("12:");
  r_val_fmt(((12.0 - led_vf) / led_if), str_buf);
  lcd.print(str_buf);
}


void loop() {
  int i;

  for (i = 0; i < 80; i++) {  // 頻繁にLCDに表示するとちらついてしまうので表示は1秒ごと程度に。
    ctrl_led();               // LCDへの表示のタイミングに関わらす、可変抵抗の値は随時PWMに反映。
    delay(10);
  }

  read_ve_vf_if();

#ifdef SERIAL_PRINT
	Serial.print("Ve1: ");
  Serial.print(ve_1);
	Serial.print(" Vf1: ");
	Serial.println(led_vf_1);
  
  Serial.print("Ve2: ");
  Serial.print(ve_2);
	Serial.print(" Vf2: ");
	Serial.println(led_vf_2);
#endif

  lcd.clear();

  if ((digitalRead(PIN_PRINT_SW_1) == HIGH) && (digitalRead(PIN_PRINT_SW_2) == HIGH)) {
    print_vf_if();
  }
  else {
    if (digitalRead(PIN_PRINT_SW_1) == LOW) {
      lcd.print("1 ");
      print_r_val(led_vf_1, led_if_1);
    }
    else {
      lcd.print("2 ");
      print_r_val(led_vf_2, led_if_2);
    }
  }
}
